#!/usr/bin/python3
'''ubuntu-repository-cache charm

This 'hooks' source constitutes all of the code for handle charm hooks.
'''

import os
import pprint
import sys

sys.path.insert(0, os.path.join(os.environ['CHARM_DIR'], 'lib'))

# pylint can't find the modules  # pylint: disable=F0401
from charmhelpers.core import hookenv    # noqa: E402
from charmhelpers.core import host       # noqa: E402
from charmhelpers.contrib import unison  # noqa: E402
from ubuntu_repository_cache import (    # noqa: E402
    metadata_sync,
    mirror,
    service,
    util,
)
# pylint: enable=F0401

HOOKS = hookenv.Hooks()
LOG = hookenv.log
SERVICE = 'ubuntu-repository-cache'
PRETTYPRINT = pprint.PrettyPrinter(indent=4)


@HOOKS.hook('debug')
def debug():
    '''Debug hook for development'''
    PRETTYPRINT.pprint(hookenv.execution_environment())


@HOOKS.hook('install.real')
def install():
    '''Perform service installation'''

    LOG('Installing %s' % SERVICE)
    service.install()


@HOOKS.hook('config-changed')
def config_changed():
    '''Handle charm configuration changes'''

    LOG('Config changed for %s' % SERVICE)
    service.render_configs()
    service.update_checks()

    # We should always do this anyway, but currently,
    # due to LP#1626318, following a reboot, the ephemeral
    # disk will be rendered immutable and read-only.
    azure_disk_warning = 'ephemeral-disk-warning'
    if host.service_available(azure_disk_warning):
        LOG('Disabling {}'.format(azure_disk_warning))
        host.service_pause(azure_disk_warning)



@HOOKS.hook('upgrade-charm')
def upgrade_charm():
    '''Handle charm updates'''

    LOG('Upgrading %s' % SERVICE)
    service.render_configs()
    service.update_checks()


@HOOKS.hook('ubuntu_repository_cache_sync')
@util.run_as_user('www-sync')
def ubuntu_repository_cache_sync():
    '''Notify clustered peers that metadata is available.

    An external system (a cron job) will trigger this hook after it has
    finished syncing metadata from upstream. This will notify the peers that
    metadata is ready.'''

    LOG('Notifying peers of metadata update')
    config = hookenv.config()
    apache_root = config['apache-root']
    apache_data = '/'.join((apache_root, 'data'))

    meta_ver = os.environ['META_VER']
    if not meta_ver:
        LOG('Leader could not gather new metadata', hookenv.WARNING)
        return

    link_next = '/'.join((apache_data, 'ubuntu_next'))

    meta_dir = os.path.realpath(link_next)

    # Push the metadata to peers
    if hookenv.is_relation_made('cluster'):
        dest = '/'.join((apache_data, meta_ver))
        link_dir = '/'.join((apache_data, 'ubuntu_active'))
        LOG('Pushing updates to peers', hookenv.DEBUG)
        source = meta_dir + '/'
        successful_peers = mirror.rsync_to_peers(
            source, dest, link_dir=link_dir)

        # Push metaversion_good canary to indicate a complete sync
        canary = '_'.join((meta_dir, 'good'))
        LOG('Pushing canary files to peers', hookenv.DEBUG)
        mirror.rsync_to_peers(canary, apache_data, peers=successful_peers)

        # Update 'meta_version' relation setting
        LOG('', hookenv.DEBUG)
        ids = hookenv.relation_ids('cluster') or []
        LOG('Updating meta_version for ids %s' % ids, hookenv.DEBUG)
        for r_id in ids:
            LOG('Setting meta_version=%s' % meta_ver, hookenv.DEBUG)
            hookenv.relation_set(relation_id=r_id, meta_version=meta_ver)

    util.update_active_symlink(os.path.basename(meta_dir))
    metadata_sync.clean_metadata_dir()
    service.start()


@HOOKS.hook('start')
def start():
    '''Start the charm's services'''

    LOG('Starting %s' % SERVICE)
    service.render_configs()

    sync_host = hookenv.config()['sync-host']
    config = hookenv.config()
    if config['sync-on-start']:
        LOG('Performing initial sync from %s' % sync_host)
        meta_ver = metadata_sync.mirror_archive(
            sync_host,
            hookenv.config()['apache-root'],
            hookenv.local_unit(),
            LOG)

        if meta_ver:
            LOG('Initial sync from {} completed ver={}.'.format(sync_host,
                                                                meta_ver))
            service.start()
        else:
            LOG('Leader could not gather new metadata', hookenv.WARNING)
    else:
        LOG('Skipping initial content sync, "sync-on-start" config '
            'option was not set')


@HOOKS.hook('stop')
def stop():
    '''Stop the charm's services'''

    LOG('Stopping %s' % SERVICE)
    service.stop()


@HOOKS.hook('website-relation-joined')
def website_relation_joined():
    '''Provide hostname/port for the website relationship'''

    LOG('Website relation joined for %s' % SERVICE)
    settings = {'hostname': hookenv.unit_private_ip(),
                'port': 80}
    hookenv.relation_set(relation_settings=settings)


@HOOKS.hook('cluster-relation-joined')
def cluster_relation_joined():
    '''Join a peer cluster relationship'''

    LOG('Cluster relation joined for %s' % SERVICE)
    unison.ssh_authorized_peers(peer_interface='cluster',
                                user='www-sync', group='www-data')
    mirror.peer_update_metadata()
    service.render_configs()


@HOOKS.hook('cluster-relation-changed')
def cluster_relation_changed():
    '''Handle peer cluster relationship changes'''

    LOG('Cluster relation changed for %s' % SERVICE)
    unison.ssh_authorized_peers(peer_interface='cluster',
                                user='www-sync', group='www-data')
    mirror.peer_update_metadata()
    service.render_configs()


@HOOKS.hook('cluster-relation-departed')
def cluster_relation_departed():
    '''Handle peer cluster relationship departures'''

    LOG('Cluster relation departed for %s' % SERVICE)
    unison.ssh_authorized_peers(peer_interface='cluster',
                                user='www-sync', group='www-data')
    mirror.peer_update_metadata()
    service.render_configs()


@HOOKS.hook('cluster-relation-broken')
def cluster_relation_broken():
    '''Leave a peer cluster relationship'''

    LOG('Cluster relation broken for %s' % SERVICE)
    service.render_configs()


@HOOKS.hook('nrpe-external-master-relation-changed')
@HOOKS.hook('local-monitors-relation-changed')
def update_checks():
    '''Update Nagios checks for all services'''

    service.update_checks()


@HOOKS.hook('leader-elected')
def leader_elected():
    """
    The current unit just was elected leader of the cluster and will be "it"
    for at least 30 seconds.

    We need to tell other units what the unit name for the leader is (the unit
    name of the currently-running hook).
    """
    hookenv.leader_set(leader_id=hookenv.local_unit())


@HOOKS.hook('leader-settings-changed')
def leader_settings_changed():
    """
    The leader settings changed. Most likely a new leader was elected and the
    leader_id was updated with leader_set.

    This simply logs the fact that it happened, for debugging purposes and to
    get a clearer picture of what happens.
    """

    leader_id = hookenv.leader_get(attribute="leader_id")
    LOG('leader-settings-changed fired. New leader_id: {}'.format(leader_id))


if __name__ == "__main__":
    # execute a hook based on the name the program is called by
    HOOKS.execute(sys.argv)
